import enum

"""
This module contains all the possible random distributions names
that can be set in each of the Space variables
"""


class ExtendedEnum(enum.Enum):
    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))


class IntegerDistributions(ExtendedEnum):
    uniform = "uniform"


class ContinuousDistributions(ExtendedEnum):
    uniform = "uniform"
    log_uniform = "log-uniform"


class CategoricalDistributions(ExtendedEnum):
    choice = "choice"



from .early_stoppers import (
    DeltaThreshold,
    ThresholdStopping,
    ConsecutiveStopping,
    TimerStopping,
)
from .loggers import ProgressBar, LogbookSaver, TensorBoard

__all__ = [
    "ProgressBar",
    "DeltaThreshold",
    "ThresholdStopping",
    "ConsecutiveStopping",
    "TimerStopping",
    "LogbookSaver",
    "TensorBoard",
]



import matplotlib.pyplot as plt
from sklearn_genetic import GAFeatureSelectionCV
from sklearn_genetic import GASearchCV
from sklearn_genetic.space import Categorical, Integer, Continuous
from sklearn.model_selection import train_test_split, StratifiedKFold
from sklearn.tree import DecisionTreeClassifier
from sklearn.datasets import load_iris
from sklearn.metrics import accuracy_score
from sklearn_genetic.mlflow_log import MLflowConfig
import numpy as np


data = load_iris()
label_names = data["target_names"]
y = data["target"]
X = data["data"]

noise = np.random.uniform(1, 4, size=(X.shape[0], 5))

X = np.hstack((X, noise))

print(X.shape)

X_train, X_test, y_train, y_test = train_test_split(
    X, y, test_size=0.33, random_state=42
)

clf = DecisionTreeClassifier()
cv = StratifiedKFold(n_splits=3, shuffle=True)


evolved_estimator = GAFeatureSelectionCV(
    estimator=clf,
    cv=cv,
    scoring="accuracy",
    population_size=30,
    generations=20,
    n_jobs=-1,
    verbose=True,
    keep_top_k=2,
    elitism=True,
)


evolved_estimator.fit(X, y)
print(evolved_estimator.best_features_)



import logging

# Check if mlflow is installed as an extra requirement
try:
    import mlflow
except ModuleNotFoundError:  # noqa
    logger = logging.getLogger(__name__)  # noqa
    logger.error("MLflow not found, pip install mlflow to use MLflowConfig")  # noqa


class MLflowConfig:
    """
    Logs each fit of hyperparameters in a running instance of mlflow: https://mlflow.org/
    """

    def __init__(
        self,
        tracking_uri,
        experiment,
        run_name,
        save_models=False,
        registry_uri=None,
        tags=None,
    ):
        """

        Parameters
        ----------
        tracking_uri: str
            Address of local or remote tracking server.
        experiment: str
            Case sensitive name of an experiment to be activated.
        run_name: str
            Name of new run (stored as a mlflow.runName tag).
        save_models: bool, default=False
            If ``True``, it will log the estimator into mlflow artifacts
        registry_uri: str, default=None
            Address of local or remote model registry server.
        tags: dict, default=None
            Dictionary of tag_name: String -> value.

        """
        self.client = mlflow.tracking.MlflowClient()
        self.tracking_uri = tracking_uri
        self.experiment = experiment
        self.run_name = run_name
        self.save_models = save_models
        self.tags = tags
        self.registry_uri = registry_uri

        mlflow.set_registry_uri(self.registry_uri)
        mlflow.set_tracking_uri(self.tracking_uri)
        mlflow.set_experiment(self.experiment)

        self.experiment_id = mlflow.get_experiment_by_name(
            self.experiment
        ).experiment_id

        if self.tags is not None:
            mlflow.set_tags(self.tags)

    def create_run(self, parameters, score, estimator):
        """
        Parameters
        ----------
        parameters: dict
            A dictionary with the keys as the hyperparameter name and the value as the current value setting
        score:
            The cross-validation score achieved by the current parameters
        estimator: estimator object
            The current sklearn estimator that is being fitted

        """

        with mlflow.start_run(
            experiment_id=self.experiment_id, nested=True, run_name=self.run_name
        ):

            for parameter, value in parameters.items():
                mlflow.log_param(key=parameter, value=value)

            mlflow.log_metric(key="score", value=score)

            if self.save_models:
                mlflow.sklearn.log_model(estimator, "model")



from abc import ABC, abstractmethod


class BaseDimension(ABC):
    """
    Base class for the space definition of data types
    """

    @abstractmethod
    def sample(self):
        """
        Sample a random value from the assigned distribution
        """

        raise NotImplementedError(
            "The sample method must be defined according each data type handler"
        )  # pragma: no cover



import logging

logger = logging.getLogger(__name__)  # noqa

# Check if seaborn is installed as an extra requirement
try:
    import seaborn as sns
except ModuleNotFoundError:  # noqa
    logger.error(
        "seaborn not found, pip install seaborn to use plots functions"
    )  # noqa
import pandas as pd
import numpy as np

from .utils import logbook_to_pandas
from .parameters import Metrics
from .space import Categorical
from .genetic_search import GAFeatureSelectionCV

"""
This module contains some useful function to explore the results of the optimization routines
"""


def plot_fitness_evolution(estimator, metric="fitness"):
    """
    Parameters
    ----------
    estimator: estimator object
        A fitted estimator from :class:`~sklearn_genetic.GASearchCV`
    metric: {"fitness", "fitness_std", "fitness_max", "fitness_min"}, default="fitness"
        Logged metric into the estimator history to plot

    Returns
    -------
    Lines plot with the fitness value in each generation

    """

    if metric not in Metrics.list():
        raise ValueError(
            f"metric must be one of {Metrics.list()}, but got {metric} instead"
        )

    sns.set_style("white")

    fitness_history = estimator.history[metric]

    palette = sns.color_palette("rocket")
    sns.set(rc={"figure.figsize": (10, 10)})

    ax = sns.lineplot(
        x=range(len(estimator)), y=fitness_history, markers=True, palette=palette
    )
    ax.set_title(f"{metric.capitalize()} average evolution over generations")

    ax.set(xlabel="generations", ylabel=f"fitness ({estimator.scoring})")
    return ax


def plot_search_space(estimator, height=2, s=25, features: list = None):
    """
    Parameters
    ----------
    estimator: estimator object
        A fitted estimator from :class:`~sklearn_genetic.GASearchCV`
    height: float, default=2
        Height of each facet
    s: float, default=5
        Size of the markers in scatter plot
    features: list, default=None
        Subset of features to plot, if ``None`` it plots all the features by default

    Returns
    -------
    Pair plot of the used hyperparameters during the search

    """

    if isinstance(estimator, GAFeatureSelectionCV):
        raise TypeError("Estimator must be a GASearchCV instance, not a GAFeatureSelectionCV instance")

    sns.set_style("white")

    df = logbook_to_pandas(estimator.logbook)
    if features:
        stats = df[features]
    else:
        variables = [*estimator.space.parameters, "score"]
        stats = df[variables]

    g = sns.PairGrid(stats, diag_sharey=False, height=height)
    g = g.map_upper(sns.scatterplot, s=s, color="r", alpha=0.2)
    g = g.map_lower(
        sns.kdeplot,
        shade=True,
        cmap=sns.color_palette("ch:s=.25,rot=-.25", as_cmap=True),
    )
    g = g.map_diag(sns.kdeplot, shade=True, palette="crest", alpha=0.2, color="red")
    return g


def noise(score):
    """
    Parameters
    ----------
    score: Series
        The `score` column from the logbook data of :class:`~sklearn_genetic.GASearchCV`

    Returns
    -------
    Noise to be added to each element of the score to avoid non-unique bin edges

    """
    score_len = len(score)
    score_std = score.std()
    noise_ratio = 1e7
    noise = (np.random.random(score_len) * score_std / noise_ratio) - (
        score_std / 2 * noise_ratio
    )
    return noise


def plot_parallel_coordinates(estimator, features: list = None):
    """
    Parameters
    ----------
    estimator: estimator object
        A fitted estimator from :class:`~sklearn_genetic.GASearchCV`
    features: list, default=None
        Subset of features to plot, if ``None`` it plots all the features by default

    Returns
    -------
    Parallel Coordinates plot of the non-categorical values

    """

    if isinstance(estimator, GAFeatureSelectionCV):
        raise TypeError("Estimator must be a GASearchCV instance, not a GAFeatureSelectionCV instance")

    df = logbook_to_pandas(estimator.logbook)
    param_grid = estimator.space.param_grid
    score = df["score"]
    if features:
        non_categorical_features = []
        for feature in features:
            if not isinstance(param_grid[feature], Categorical):
                non_categorical_features.append(feature)
            else:
                logger.warning(
                    "`%s` is Categorical variable! It was dropped from the plot feature list",
                    feature,
                )
        stats = df[non_categorical_features]
    else:
        non_categorical_variables = []
        for variable, var_type in param_grid.items():
            if not isinstance(var_type, Categorical):
                non_categorical_variables.append(variable)
        non_categorical_variables.append("score")
        stats = df[non_categorical_variables]

    stats["score_quartile"] = pd.qcut(score + noise(score), 4, labels=[1, 2, 3, 4])
    g = pd.plotting.parallel_coordinates(
        stats, "score_quartile", color=("#8E8E8D", "#4ECDC4", "#C7F464", "#FF0000")
    )

    return g



from .genetic_search import GASearchCV, GAFeatureSelectionCV

from .callbacks import (
    ThresholdStopping,
    ConsecutiveStopping,
    DeltaThreshold,
    LogbookSaver,
)


from ._version import __version__

__all__ = [
    "GASearchCV",
    "GAFeatureSelectionCV",
    "ThresholdStopping",
    "ConsecutiveStopping",
    "DeltaThreshold",
    "LogbookSaver",
    "__version__",
]



from datetime import datetime

from .validations import check_stats
from .base import BaseCallback


class ThresholdStopping(BaseCallback):
    """
    Stop the optimization if the metric from
    cross validation score is greater or equals than the define threshold
    """

    def __init__(self, threshold, metric="fitness"):
        """
        Parameters
        ----------
        threshold: float, default=None
            Threshold to compare against the current cross validation average score and determine if
            the optimization process must stop
        metric: {'fitness', 'fitness_std', 'fitness_max', 'fitness_min'}, default ='fitness'
            Name of the metric inside 'record' logged in each iteration
        """

        check_stats(metric)

        self.threshold = threshold
        self.metric = metric

    def on_step(self, record=None, logbook=None, estimator=None):
        stat = None
        if record is not None:
            stat = record[self.metric]
        elif logbook is not None:
            # Get the last metric value
            stat = logbook.select(self.metric)[-1]

        else:
            raise ValueError(
                "At least one of record or logbook parameters must be provided"
            )

        if stat is not None and stat >= self.threshold:
            print(f"INFO: {self.__class__.__name__} callback met its criteria")
            return True
        return False


class ConsecutiveStopping(BaseCallback):
    """
    Stop the optimization if the current metric value is no greater that at least one metric from the last N generations
    """

    def __init__(self, generations, metric="fitness"):
        """
        Parameters
        ----------
        generations: int, default=None
            Number of current generations to compare against current generation
        metric: {'fitness', 'fitness_std', 'fitness_max', 'fitness_min'}, default ='fitness'
            Name of the metric inside 'record' logged in each iteration
        """

        check_stats(metric)

        self.generations = generations
        self.metric = metric

    def on_step(self, record=None, logbook=None, estimator=None):
        if logbook is not None:
            if len(logbook) <= self.generations:
                return False

            if record is not None:
                current_stat = record[self.metric]
            else:
                current_stat = logbook.select(self.metric)[-1]

            # Compare the current metric with the last |generations| metrics
            stats = logbook.select(self.metric)[(-self.generations - 1) : -1]

            if all(stat >= current_stat for stat in stats):
                print(f"INFO: {self.__class__.__name__} callback met its criteria")
                return True
            return False

        else:
            raise ValueError("logbook parameter must be provided")


class DeltaThreshold(BaseCallback):
    """
    Stops the optimization if the absolute difference between the maximum and minimum value from the last N generations
    is less or equals to a threshold.
    The threshold gets evaluated after the number of generations specified is reached.
    """

    def __init__(self, threshold, generations=2, metric: str = "fitness"):
        """
        Parameters
        ----------
        threshold: float, default=None
            Threshold to compare the differences between cross-validation scores.
        generations: int, default=2
            Number of generations to compare, includes the current generation.
        metric: {'fitness', 'fitness_std', 'fitness_max', 'fitness_min'}, default ='fitness'
            Name of the metric inside 'record' logged in each iteration.
        """

        check_stats(metric)

        self.threshold = threshold
        self.generations = generations
        self.metric = metric

    def on_step(self, record=None, logbook=None, estimator=None):
        if logbook is not None:
            if len(logbook) < self.generations:
                return False

            stats = logbook.select(self.metric)[(-self.generations) :]

            diff = abs(max(stats) - min(stats))

            if diff <= self.threshold:
                print(f"INFO: {self.__class__.__name__} callback met its criteria")
                return True
            return False

        else:
            raise ValueError("logbook parameter must be provided")


class TimerStopping(BaseCallback):
    """
    Stops the optimization process if a limit training time has been elapsed.
    This time is checked after each generation fit
    """

    def __init__(self, total_seconds):
        """
        Parameters
        ----------
        total_seconds: int
            Total time in seconds that the estimator is allowed to fit
        """
        self.initial_training_time = None
        self.total_seconds = total_seconds

    def on_start(self, estimator=None):
        self.initial_training_time = datetime.utcnow()

    def on_step(self, record=None, logbook=None, estimator=None):
        current_time = datetime.utcnow()
        difference = current_time - self.initial_training_time
        difference_seconds = difference.total_seconds()

        if difference_seconds >= self.total_seconds:
            print(f"INFO: {self.__class__.__name__} callback met its criteria")
            return True
        return False



from ..parameters import Metrics, CallbackMethods
from .base import BaseCallback


def check_stats(metric):
    if metric not in Metrics.list():
        raise ValueError(
            f"metric must be one of {Metrics.list()}, but got {metric} instead"
        )


def check_callback(callback):
    """
    Check if callback is a callable or a list of callables.
    """
    if callback is not None:
        if isinstance(callback, BaseCallback):
            return [callback]

        elif isinstance(callback, list) and all(
            [isinstance(c, BaseCallback) for c in callback]
        ):
            return callback

        else:
            raise ValueError(
                "callback should be either a class or a list of classes with inheritance from "
                "callbacks.base.BaseCallback"
            )
    else:
        return []


def eval_callbacks(callbacks, record, logbook, estimator, method):
    """Evaluate list of callbacks on result.
    Parameters
    ----------
    callbacks : list of callables
        Callbacks to evaluate.
    record : logbook record
    logbook: logbook object
            Current stream logbook with the stats required
    estimator: :class:`~sklearn_genetic.GASearchCV`, default = None
        Estimator that is being optimized
    method: {'on_start', 'on_step', 'on_end'}
        The method to be called from the callback

    Returns
    -------
    decision : bool
        Decision of the callbacks whether or not to keep optimizing
    """

    if method not in CallbackMethods.list():
        raise ValueError(
            f"The callback method must be one of {CallbackMethods.list()}, but got {method} instead"
        )

    stop = False
    decision = None

    if callbacks:
        for callback in callbacks:
            callback_method = getattr(callback, method)

            if method == CallbackMethods.on_start.value:
                decision = callback_method(estimator)

            elif method == CallbackMethods.on_step.value:
                decision = callback_method(record, logbook, estimator)

            elif method == CallbackMethods.on_end.value:
                decision = callback_method(logbook, estimator)

            if decision is not None:
                stop = stop or decision

    return stop
